import time
import random
import asyncio
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Generator, AsyncGenerator
from functools import cached_property
from typing import Any, overload, Literal

import httpx
from openai import OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI
from openai._types import Headers, Query, Body
from openai.types.shared_params.metadata import Metadata
from openai.types.chat.completion_create_params import ResponseFormat
from openai.types.chat.chat_completion_modality import ChatCompletionModality
from openai.types.chat.chat_completion_audio_param import ChatCompletionAudioParam
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from openai.types.chat.chat_completion_stream_options_param import ChatCompletionStreamOptionsParam
from openai.types.chat.chat_completion_prediction_content_param import ChatCompletionPredictionContentParam

from anthropic import (
    Anthropic,
    AsyncAnthropic,
    AnthropicVertex,
    AsyncAnthropicVertex,
    AnthropicBedrock,
    AsyncAnthropicBedrock,
)
from anthropic.types.thinking_config_param import ThinkingConfigParam

from ..settings import Settings, normalize_settings
from ..types import defaults as defs
from ..types.settings import SettingsDict
from ..types.enums import ContextLengthControlType, BackendType
from ..types.llm_parameters import (
    NotGiven,
    NOT_GIVEN,
    ToolParam,
    ToolChoice,
    OpenAINotGiven,
    EndpointSetting,
    ChatCompletionMessage,
    ChatCompletionDeltaMessage,
)
from ..utilities.rate_limiter import SyncMemoryRateLimiter, SyncRedisRateLimiter, SyncDiskCacheRateLimiter
from ..utilities.rate_limiter import AsyncMemoryRateLimiter, AsyncRedisRateLimiter, AsyncDiskCacheRateLimiter


class BaseChatClient(ABC):
    DEFAULT_MODEL: str
    BACKEND_NAME: BackendType

    def __init__(
        self,
        model: str = "",
        stream: bool = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        context_length_control: ContextLengthControlType = defs.CONTEXT_LENGTH_CONTROL,
        random_endpoint: bool = True,
        endpoint_id: str = "",
        http_client: httpx.Client | None = None,
        backend_name: str | None = None,
        settings: Settings | SettingsDict | None = None,  # Use default settings if not provided
    ):
        self.model = model or self.DEFAULT_MODEL
        self.stream = stream
        self.temperature = temperature
        self.context_length_control = context_length_control
        self.random_endpoint = random_endpoint
        self.endpoint_id = endpoint_id
        self.http_client = http_client

        if backend_name is not None:
            self.backend_name = BackendType(backend_name)
        else:
            self.backend_name = self.BACKEND_NAME

        self.settings: Settings = normalize_settings(settings)

        self.backend_settings = self.settings.get_backend(self.backend_name)

        self.rate_limiter = self._init_rate_limiter()
        self.active_requests = defaultdict(int)
        self.rpm = None
        self.tpm = None
        self.concurrent_requests = None

        if endpoint_id:
            self.endpoint_id = endpoint_id
            self.random_endpoint = False
            self.endpoint = self.settings.get_endpoint(self.endpoint_id)

    def _init_rate_limiter(self):
        if not self.settings.rate_limit:
            return None
        if not self.settings.rate_limit.enabled:
            return None

        if self.settings.rate_limit.backend == "memory":
            return SyncMemoryRateLimiter()
        elif self.settings.rate_limit.backend == "redis":
            if not self.settings.rate_limit.redis:
                raise ValueError("Redis settings must be provided if Redis backend is selected.")
            return SyncRedisRateLimiter(
                host=self.settings.rate_limit.redis.host,
                port=self.settings.rate_limit.redis.port,
                db=self.settings.rate_limit.redis.db,
            )
        elif self.settings.rate_limit.backend == "diskcache":
            if not self.settings.rate_limit.diskcache:
                raise ValueError("Diskcache settings must be provided if Diskcache backend is selected.")
            return SyncDiskCacheRateLimiter(
                cache_dir=self.settings.rate_limit.diskcache.cache_dir,
            )
        return None

    def _acquire_rate_limit(self, endpoint: EndpointSetting | None, model: str, messages: list):
        if endpoint is None:
            return

        key = f"{endpoint.id}:{model}"

        # Get rate limit parameters
        # Priority: parameters in model.endpoints > parameters in endpoint > default parameters
        rpm = self.rpm or endpoint.rpm or (self.settings.rate_limit.default_rpm if self.settings.rate_limit else 60)
        tpm = self.tpm or endpoint.tpm or (self.settings.rate_limit.default_tpm if self.settings.rate_limit else 1000000)

        while self.rate_limiter:
            allowed, wait_time = self.rate_limiter.check_limit(key, rpm, tpm, self._estimate_request_tokens(messages))
            if allowed:
                break
            time.sleep(wait_time)

    def _estimate_request_tokens(self, messages: list) -> int:
        """Roughly estimate the number of tokens in the request.
        Uses higher ratio for CJK characters which typically consume more tokens.
        """
        tokens = 0
        for message in messages:
            content = message.get("content", "")
            if isinstance(content, str):
                cjk_chars = sum(1 for c in content if '\u4e00' <= c <= '\u9fff' or '\u3000' <= c <= '\u303f' or '\uff00' <= c <= '\uffef')
                ascii_chars = len(content) - cjk_chars
                tokens += int(cjk_chars * 1.5 + ascii_chars * 0.25)
        return max(tokens, 1)

    def _get_available_endpoints(self, model_endpoints: list) -> list:
        """Get list of available (enabled) endpoints for the model"""
        available_endpoints = []
        for endpoint_option in model_endpoints:
            if isinstance(endpoint_option, dict):
                endpoint_id = endpoint_option["endpoint_id"]
                try:
                    endpoint = self.settings.get_endpoint(endpoint_id)
                    if endpoint.enabled:
                        available_endpoints.append(endpoint_option)
                except ValueError:
                    continue
            else:
                try:
                    endpoint = self.settings.get_endpoint(endpoint_option)
                    if endpoint.enabled:
                        available_endpoints.append(endpoint_option)
                except ValueError:
                    continue
        return available_endpoints

    def set_model_id_by_endpoint_id(self, endpoint_id: str):
        for endpoint_option in self.backend_settings.models[self.model].endpoints:
            if isinstance(endpoint_option, dict) and endpoint_id == endpoint_option["endpoint_id"]:
                self.model_id = endpoint_option["model_id"]
                break
        return self.model_id

    def _set_endpoint(self):
        if self.endpoint is None:
            if self.random_endpoint:
                available_endpoints = self._get_available_endpoints(self.backend_settings.models[self.model].endpoints)
                if not available_endpoints:
                    raise ValueError(f"No enabled endpoints available for model {self.model}")

                endpoint = random.choice(available_endpoints)
                if isinstance(endpoint, dict):
                    self.endpoint_id = endpoint["endpoint_id"]
                    self.model_id = endpoint["model_id"]
                    self.rpm = endpoint.get("rpm", None)
                    self.tpm = endpoint.get("tpm", None)
                    self.concurrent_requests = endpoint.get("concurrent_requests", None)
                else:
                    self.endpoint_id = endpoint
                self.endpoint = self.settings.get_endpoint(self.endpoint_id)
            else:
                self.endpoint = self.settings.get_endpoint(self.endpoint_id)
                if not self.endpoint.enabled:
                    raise ValueError(f"Endpoint {self.endpoint_id} is disabled")
                self.set_model_id_by_endpoint_id(self.endpoint_id)
        elif isinstance(self.endpoint, EndpointSetting):
            if not self.endpoint.enabled:
                raise ValueError(f"Endpoint {self.endpoint.id} is disabled")
            self.endpoint_id = self.endpoint.id
            self.set_model_id_by_endpoint_id(self.endpoint_id)
        else:
            raise ValueError("Invalid endpoint")

        return self.endpoint, self.model_id

    @cached_property
    @abstractmethod
    def raw_client(
        self,
    ) -> OpenAI | AzureOpenAI | Anthropic | AnthropicVertex | AnthropicBedrock | httpx.Client | None:
        pass

    @overload
    @abstractmethod
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage:
        pass

    @overload
    @abstractmethod
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[True],
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> Generator[ChatCompletionDeltaMessage, Any, None]:
        pass

    @overload
    @abstractmethod
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: bool,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage | Generator[ChatCompletionDeltaMessage, Any, None]:
        pass

    @abstractmethod
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] | Literal[True] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage | Generator[ChatCompletionDeltaMessage, Any, None]:
        pass

    def create_stream(
        self,
        *,
        messages: list,
        model: str | None = None,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> Generator[ChatCompletionDeltaMessage, Any, None]:
        return self.create_completion(
            messages=messages,
            model=model,
            stream=True,
            temperature=temperature,
            max_tokens=max_tokens,
            tools=tools,
            tool_choice=tool_choice,
            response_format=response_format,
            stream_options=stream_options,
            top_p=top_p,
            skip_cutoff=skip_cutoff,
            audio=audio,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            logprobs=logprobs,
            max_completion_tokens=max_completion_tokens,
            metadata=metadata,
            modalities=modalities,
            n=n,
            parallel_tool_calls=parallel_tool_calls,
            prediction=prediction,
            presence_penalty=presence_penalty,
            reasoning_effort=reasoning_effort,
            thinking=thinking,
            seed=seed,
            service_tier=service_tier,
            stop=stop,
            store=store,
            top_logprobs=top_logprobs,
            user=user,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
            timeout=timeout,
        )

    def model_list(self):
        _raw_client = self.raw_client
        if isinstance(_raw_client, OpenAI | AzureOpenAI):
            return _raw_client.models.list().model_dump()
        elif isinstance(_raw_client, Anthropic):
            return _raw_client.models.list(limit=1000).model_dump()
        else:
            raise ValueError(f"Unsupported client type: {type(_raw_client)}")


class BaseAsyncChatClient(ABC):
    DEFAULT_MODEL: str
    BACKEND_NAME: BackendType

    def __init__(
        self,
        model: str = "",
        stream: bool = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        context_length_control: ContextLengthControlType = defs.CONTEXT_LENGTH_CONTROL,
        random_endpoint: bool = True,
        endpoint_id: str = "",
        http_client: httpx.AsyncClient | None = None,
        backend_name: str | None = None,
        settings: Settings | SettingsDict | None = None,  # Use default settings if not provided
    ):
        self.model = model or self.DEFAULT_MODEL
        self.stream = stream
        self.temperature = temperature
        self.context_length_control = context_length_control
        self.random_endpoint = random_endpoint
        self.endpoint_id = endpoint_id
        self.http_client = http_client

        if backend_name is not None:
            self.backend_name = BackendType(backend_name)
        else:
            self.backend_name = self.BACKEND_NAME

        self.settings: Settings = normalize_settings(settings)

        self.backend_settings = self.settings.get_backend(self.backend_name)

        self.rate_limiter = self._init_rate_limiter()
        self.active_requests = defaultdict(int)
        self.rpm = None
        self.tpm = None
        self.concurrent_requests = None

        if endpoint_id:
            self.endpoint_id = endpoint_id
            self.random_endpoint = False
            self.endpoint = self.settings.get_endpoint(self.endpoint_id)

    def _init_rate_limiter(self):
        if not self.settings.rate_limit:
            return None
        if not self.settings.rate_limit.enabled:
            return None

        if self.settings.rate_limit.backend == "memory":
            return AsyncMemoryRateLimiter()
        elif self.settings.rate_limit.backend == "redis":
            if not self.settings.rate_limit.redis:
                raise ValueError("Redis settings must be provided if Redis backend is selected.")
            return AsyncRedisRateLimiter(
                host=self.settings.rate_limit.redis.host,
                port=self.settings.rate_limit.redis.port,
                db=self.settings.rate_limit.redis.db,
            )
        elif self.settings.rate_limit.backend == "diskcache":
            if not self.settings.rate_limit.diskcache:
                raise ValueError("Diskcache settings must be provided if Diskcache backend is selected.")
            return AsyncDiskCacheRateLimiter(
                cache_dir=self.settings.rate_limit.diskcache.cache_dir,
            )
        return None

    async def _acquire_rate_limit(self, endpoint: EndpointSetting | None, model: str, messages: list):
        if endpoint is None:
            return

        key = f"{endpoint.id}:{model}"

        # Get rate limit parameters
        # Priority: parameters in model.endpoints > parameters in endpoint > default parameters
        rpm = self.rpm or endpoint.rpm or (self.settings.rate_limit.default_rpm if self.settings.rate_limit else 60)
        tpm = self.tpm or endpoint.tpm or (self.settings.rate_limit.default_tpm if self.settings.rate_limit else 1000000)

        while self.rate_limiter:
            allowed, wait_time = await self.rate_limiter.check_limit(key, rpm, tpm, self._estimate_request_tokens(messages))
            if allowed:
                break
            await asyncio.sleep(wait_time)

    def _estimate_request_tokens(self, messages: list) -> int:
        """Roughly estimate the number of tokens in the request.
        Uses higher ratio for CJK characters which typically consume more tokens.
        """
        tokens = 0
        for message in messages:
            content = message.get("content", "")
            if isinstance(content, str):
                cjk_chars = sum(1 for c in content if '\u4e00' <= c <= '\u9fff' or '\u3000' <= c <= '\u303f' or '\uff00' <= c <= '\uffef')
                ascii_chars = len(content) - cjk_chars
                tokens += int(cjk_chars * 1.5 + ascii_chars * 0.25)
        return max(tokens, 1)

    def _get_available_endpoints(self, model_endpoints: list) -> list:
        """Get list of available (enabled) endpoints for the model"""
        available_endpoints = []
        for endpoint_option in model_endpoints:
            if isinstance(endpoint_option, dict):
                # For endpoint with specific config, check if the endpoint is enabled
                endpoint_id = endpoint_option["endpoint_id"]
                try:
                    endpoint = self.settings.get_endpoint(endpoint_id)
                    if endpoint.enabled:
                        available_endpoints.append(endpoint_option)
                except ValueError:
                    # Endpoint not found, skip it
                    continue
            else:
                # For simple endpoint ID string, check if the endpoint is enabled
                try:
                    endpoint = self.settings.get_endpoint(endpoint_option)
                    if endpoint.enabled:
                        available_endpoints.append(endpoint_option)
                except ValueError:
                    # Endpoint not found, skip it
                    continue
        return available_endpoints

    def set_model_id_by_endpoint_id(self, endpoint_id: str):
        for endpoint_option in self.backend_settings.models[self.model].endpoints:
            if isinstance(endpoint_option, dict) and endpoint_id == endpoint_option["endpoint_id"]:
                self.model_id = endpoint_option["model_id"]
                break
        return self.model_id

    def _set_endpoint(self):
        if self.endpoint is None:
            if self.random_endpoint:
                # Get available (enabled) endpoints
                available_endpoints = self._get_available_endpoints(self.backend_settings.models[self.model].endpoints)
                if not available_endpoints:
                    raise ValueError(f"No enabled endpoints available for model {self.model}")

                endpoint = random.choice(available_endpoints)
                if isinstance(endpoint, dict):
                    self.endpoint_id = endpoint["endpoint_id"]
                    self.model_id = endpoint["model_id"]
                    self.rpm = endpoint.get("rpm", None)
                    self.tpm = endpoint.get("tpm", None)
                    self.concurrent_requests = endpoint.get("concurrent_requests", None)
                else:
                    self.endpoint_id = endpoint
                self.endpoint = self.settings.get_endpoint(self.endpoint_id)
            else:
                self.endpoint = self.settings.get_endpoint(self.endpoint_id)
                # Check if the specified endpoint is enabled
                if not self.endpoint.enabled:
                    raise ValueError(f"Endpoint {self.endpoint_id} is disabled")
                self.set_model_id_by_endpoint_id(self.endpoint_id)
        elif isinstance(self.endpoint, EndpointSetting):
            # Check if the endpoint is enabled
            if not self.endpoint.enabled:
                raise ValueError(f"Endpoint {self.endpoint.id} is disabled")
            self.endpoint_id = self.endpoint.id
            self.set_model_id_by_endpoint_id(self.endpoint_id)
        else:
            raise ValueError("Invalid endpoint")

        return self.endpoint, self.model_id

    @cached_property
    @abstractmethod
    def raw_client(
        self,
    ) -> AsyncOpenAI | AsyncAzureOpenAI | AsyncAnthropic | AsyncAnthropicVertex | AsyncAnthropicBedrock | httpx.AsyncClient | None:
        pass

    @overload
    @abstractmethod
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage:
        pass

    @overload
    @abstractmethod
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[True],
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        pass

    @overload
    @abstractmethod
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: bool,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage | AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        pass

    @abstractmethod
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] | Literal[True] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage | AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        pass

    async def create_stream(
        self,
        *,
        messages: list,
        model: str | None = None,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        return await self.create_completion(
            messages=messages,
            model=model,
            stream=True,
            temperature=temperature,
            max_tokens=max_tokens,
            tools=tools,
            tool_choice=tool_choice,
            response_format=response_format,
            stream_options=stream_options,
            top_p=top_p,
            skip_cutoff=skip_cutoff,
            audio=audio,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            logprobs=logprobs,
            max_completion_tokens=max_completion_tokens,
            metadata=metadata,
            modalities=modalities,
            n=n,
            parallel_tool_calls=parallel_tool_calls,
            prediction=prediction,
            presence_penalty=presence_penalty,
            reasoning_effort=reasoning_effort,
            thinking=thinking,
            seed=seed,
            service_tier=service_tier,
            stop=stop,
            store=store,
            top_logprobs=top_logprobs,
            user=user,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
            timeout=timeout,
        )

    async def model_list(self):
        _raw_client = self.raw_client
        if isinstance(_raw_client, AsyncOpenAI | AsyncAzureOpenAI):
            return (await _raw_client.models.list()).model_dump()
        elif isinstance(_raw_client, AsyncAnthropic):
            return (await _raw_client.models.list(limit=1000)).model_dump()
        else:
            raise ValueError(f"Unsupported client type: {type(_raw_client)}")



import json
from typing import Any, cast

from anthropic.types import (
    RawContentBlockDeltaEvent,
    RawContentBlockStartEvent,
    RawContentBlockStopEvent,
    RawMessageDeltaEvent,
    RawMessageStartEvent,
    RawMessageStopEvent,
    TextBlock,
    ThinkingBlock,
    ToolUseBlock,
)
from openai.types.completion_usage import PromptTokensDetails

from ..types.llm_parameters import ChatCompletionDeltaMessage, ChatCompletionMessage, Usage
from .tool_call_parser import refactor_tool_calls


def adapt_response_api_stream_event(event: Any, final_tool_calls: dict[int, dict[str, Any]], is_gemini3: bool) -> tuple[list[ChatCompletionDeltaMessage], Usage | None]:
    """Convert a Responses API stream event into ChatCompletionDeltaMessage objects."""
    event_type = event.type

    if event_type == "response.output_text.delta":
        if event.delta:
            return [ChatCompletionDeltaMessage(content=event.delta)], None
        return [], None

    if event_type == "response.output_item.added":
        item = event.item
        output_index = event.output_index
        if item and item.type == "function_call" and output_index is not None:
            extra_content = getattr(item, "extra_content", None)
            final_tool_calls[output_index] = {
                "id": item.id,
                "call_id": item.call_id,
                "name": item.name,
                "arguments": item.arguments or "",
                "extra_content": extra_content,
            }
            return [
                ChatCompletionDeltaMessage(
                    tool_calls=cast(
                        Any,
                        [
                            {
                                "index": output_index,
                                "id": item.id,
                                "type": "function",
                                "function": {
                                    "name": item.name,
                                    "arguments": "",
                                },
                            }
                        ],
                    ),
                )
            ], None
        return [], None

    if event_type == "response.function_call_arguments.delta":
        output_index = event.output_index
        delta = event.delta
        if output_index is not None and delta is not None:
            if output_index in final_tool_calls:
                final_tool_calls[output_index]["arguments"] += delta
                call_id = final_tool_calls[output_index].get("id")
            else:
                call_id = None
            return [
                ChatCompletionDeltaMessage(
                    tool_calls=cast(
                        Any,
                        [
                            {
                                "index": output_index,
                                "id": call_id,
                                "type": "function",
                                "function": {"arguments": delta},
                            }
                        ],
                    ),
                )
            ], None
        return [], None

    if event_type == "response.function_call_arguments.done":
        output_index = event.output_index
        if output_index is not None and output_index in final_tool_calls:
            tool_call = final_tool_calls[output_index]
            raw_content = None
            if is_gemini3:
                extra_content = tool_call.get("extra_content")
                if isinstance(extra_content, dict) and extra_content.get("google"):
                    raw_content = {"google": extra_content["google"]}
            return [
                ChatCompletionDeltaMessage(
                    tool_calls=cast(
                        Any,
                        [
                            {
                                "index": output_index,
                                "id": tool_call.get("id"),
                                "type": "function",
                                "function": {
                                    "name": tool_call.get("name"),
                                    "arguments": tool_call.get("arguments", ""),
                                },
                            }
                        ],
                    ),
                    raw_content=raw_content,
                )
            ], None
        return [], None

    if event_type == "response.output_item.done":
        item = event.item
        output_index = event.output_index
        if item and item.type == "function_call" and output_index is not None:
            return [
                ChatCompletionDeltaMessage(
                    tool_calls=cast(
                        Any,
                        [
                            {
                                "index": output_index,
                                "id": item.id,
                                "type": "function",
                                "function": {
                                    "name": item.name,
                                    "arguments": item.arguments,
                                },
                            }
                        ],
                    ),
                )
            ], None
        return [], None

    if event_type == "response.completed":
        final_resp = event.response
        if final_resp and final_resp.usage:
            usage = final_resp.usage
            return [], Usage(
                completion_tokens=usage.output_tokens or 0,
                prompt_tokens=usage.input_tokens or 0,
                total_tokens=(usage.input_tokens or 0) + (usage.output_tokens or 0),
            )
        return [], None

    if event_type in ("response.error", "error"):
        raise RuntimeError(f"Responses stream error: {event.error}")

    return [], None


def init_anthropic_stream_state() -> dict[str, Any]:
    return {
        "content": "",
        "reasoning_content": "",
        "usage": {},
        "tool_calls": [],
        "raw_content": [],
    }


def _last_raw_content(stream_state: dict[str, Any], block_type: str) -> dict[str, Any] | None:
    for i in range(len(stream_state["raw_content"]) - 1, -1, -1):
        item = stream_state["raw_content"][i]
        if item["type"] == block_type:
            return cast(dict[str, Any], item)
    return None


def adapt_anthropic_stream_event(chunk: Any, stream_state: dict[str, Any]) -> ChatCompletionDeltaMessage | None:
    message: dict[str, Any] = {"content": "", "tool_calls": []}

    if isinstance(chunk, RawMessageStartEvent):
        cache_read_tokens = getattr(chunk.message.usage, "cache_read_input_tokens", 0) or 0
        prompt_tokens = chunk.message.usage.input_tokens + cache_read_tokens
        usage_data: dict[str, Any] = {"prompt_tokens": prompt_tokens}
        if cache_read_tokens:
            usage_data["prompt_tokens_details"] = {"cached_tokens": cache_read_tokens}
        stream_state["usage"] = usage_data
        return None

    if isinstance(chunk, RawContentBlockStartEvent):
        content_block = chunk.content_block.model_dump()
        stream_state["raw_content"].append(content_block)
        if chunk.content_block.type == "tool_use":
            stream_state["tool_calls"] = message["tool_calls"] = [
                {
                    "index": 0,
                    "id": chunk.content_block.id,
                    "function": {
                        "arguments": "",
                        "name": chunk.content_block.name,
                    },
                    "type": "function",
                }
            ]
        elif chunk.content_block.type == "text":
            message["content"] = chunk.content_block.text
        elif chunk.content_block.type == "thinking":
            message["reasoning_content"] = chunk.content_block.thinking
        message["raw_content"] = content_block
        return ChatCompletionDeltaMessage(**message)

    if isinstance(chunk, RawContentBlockDeltaEvent):
        if chunk.delta.type == "text_delta":
            message["content"] = chunk.delta.text
            stream_state["content"] += chunk.delta.text
            text_block = _last_raw_content(stream_state, "text")
            if text_block is not None:
                text_block["text"] += chunk.delta.text
        elif chunk.delta.type == "thinking_delta":
            message["reasoning_content"] = chunk.delta.thinking
            stream_state["reasoning_content"] += chunk.delta.thinking
            thinking_block = _last_raw_content(stream_state, "thinking")
            if thinking_block is not None:
                thinking_block["thinking"] += chunk.delta.thinking
        elif chunk.delta.type == "signature_delta":
            thinking_block = _last_raw_content(stream_state, "thinking")
            if thinking_block is not None:
                if "signature" not in thinking_block:
                    thinking_block["signature"] = ""
                thinking_block["signature"] += chunk.delta.signature
        elif chunk.delta.type == "citations_delta":
            citation_data = chunk.delta.citation.model_dump()
            raw_content = _last_raw_content(stream_state, stream_state["raw_content"][-1]["type"]) if stream_state["raw_content"] else None
            if raw_content is not None:
                if "citations" not in raw_content:
                    raw_content["citations"] = []
                raw_content["citations"].append(citation_data)
        elif chunk.delta.type == "input_json_delta" and stream_state["tool_calls"]:
            stream_state["tool_calls"][0]["function"]["arguments"] += chunk.delta.partial_json
            message["tool_calls"] = [
                {
                    "index": 0,
                    "id": stream_state["tool_calls"][0]["id"],
                    "function": {
                        "arguments": chunk.delta.partial_json,
                        "name": stream_state["tool_calls"][0]["function"]["name"],
                    },
                    "type": "function",
                }
            ]
            tool_use_block = _last_raw_content(stream_state, "tool_use")
            if tool_use_block is not None:
                if "input" not in tool_use_block:
                    tool_use_block["input"] = {}
                try:
                    if stream_state["tool_calls"][0]["function"]["arguments"]:
                        tool_use_block["input"] = json.loads(stream_state["tool_calls"][0]["function"]["arguments"])
                    else:
                        tool_use_block["input"] = {}
                except json.JSONDecodeError:
                    pass
        elif chunk.delta.type == "redacted_thinking_delta":
            redacted_block = _last_raw_content(stream_state, "redacted_thinking")
            if redacted_block is not None:
                if "data" not in redacted_block:
                    redacted_block["data"] = ""
                redacted_block["data"] += chunk.delta.data

        message["raw_content"] = chunk.delta.model_dump()
        return ChatCompletionDeltaMessage(**message)

    if isinstance(chunk, RawMessageDeltaEvent):
        stream_state["usage"]["completion_tokens"] = chunk.usage.output_tokens
        stream_state["usage"]["total_tokens"] = stream_state["usage"]["prompt_tokens"] + stream_state["usage"]["completion_tokens"]
        usage_kwargs: dict[str, Any] = {
            "prompt_tokens": stream_state["usage"]["prompt_tokens"],
            "completion_tokens": stream_state["usage"]["completion_tokens"],
            "total_tokens": stream_state["usage"]["total_tokens"],
        }
        if "prompt_tokens_details" in stream_state["usage"]:
            usage_kwargs["prompt_tokens_details"] = PromptTokensDetails(cached_tokens=stream_state["usage"]["prompt_tokens_details"].get("cached_tokens", 0))
        return ChatCompletionDeltaMessage(usage=Usage(**usage_kwargs))

    if isinstance(chunk, RawMessageStopEvent | RawContentBlockStopEvent):
        return None

    return None


def build_anthropic_completion_message(content_blocks: list[Any], usage: Any) -> ChatCompletionMessage:
    result: dict[str, Any] = {
        "content": "",
        "reasoning_content": "",
        "raw_content": [content_block.model_dump() for content_block in content_blocks],
        "usage": {
            "prompt_tokens": usage.input_tokens + usage.cache_read_input_tokens if usage.cache_read_input_tokens else usage.input_tokens,
            "completion_tokens": usage.output_tokens,
            "total_tokens": usage.input_tokens + usage.output_tokens,
            "prompt_tokens_details": {
                "cached_tokens": usage.cache_read_input_tokens,
            },
        },
    }
    tool_calls = []
    for content_block in content_blocks:
        if isinstance(content_block, TextBlock):
            result["content"] += content_block.text
        elif isinstance(content_block, ThinkingBlock):
            result["reasoning_content"] = content_block.thinking
        elif isinstance(content_block, ToolUseBlock):
            tool_calls.append(content_block.model_dump())

    if tool_calls:
        result["tool_calls"] = refactor_tool_calls(tool_calls)

    return ChatCompletionMessage(**result)



# @Author: Bi Ying
# @Date:   2024-07-26 14:48:55
import json
from functools import cached_property
from collections.abc import Generator, AsyncGenerator, Iterable
from typing import Any, TYPE_CHECKING, overload, Literal, cast

import httpx
from openai import OpenAI, AsyncOpenAI, AzureOpenAI, AsyncAzureOpenAI
from openai._types import Headers, Query, Body
from openai.types.shared_params.metadata import Metadata
from openai.types.completion_usage import PromptTokensDetails
from openai.types.chat.completion_create_params import ResponseFormat
from openai.types.chat.chat_completion_modality import ChatCompletionModality
from openai.types.chat.chat_completion_audio_param import ChatCompletionAudioParam
from openai.types.chat.chat_completion_reasoning_effort import ChatCompletionReasoningEffort
from openai.types.chat.chat_completion_stream_options_param import ChatCompletionStreamOptionsParam
from openai.types.chat.chat_completion_prediction_content_param import ChatCompletionPredictionContentParam
from anthropic.types.thinking_config_param import ThinkingConfigParam

from .base_client import BaseChatClient, BaseAsyncChatClient
from .utils import (
    cutoff_messages,
    get_message_token_counts,
    ToolCallContentProcessor,
    generate_tool_use_system_prompt,
)
from .message_normalizer import (
    extract_reasoning_tagged_content,
    get_thinking_tags,
    is_gemini_3_model,
    process_thinking_content,
)
from .stream_event_adapter import adapt_response_api_stream_event
from .tool_call_parser import tool_schema_for_response_api
from ..types import defaults as defs
from ..utilities.gcp_token import get_token_with_cache
from ..types.enums import ContextLengthControlType, BackendType
from ..types.llm_parameters import (
    NotGiven,
    NOT_GIVEN,
    OPENAI_NOT_GIVEN,
    ToolParam,
    ToolChoice,
    OpenAINotGiven,
    AnthropicNotGiven,
    Usage,
    ChatCompletionMessage,
    ChatCompletionDeltaMessage,
)

if TYPE_CHECKING:
    from ..settings import Settings
    from ..types.settings import SettingsDict


class OpenAICompatibleChatClient(BaseChatClient):
    DEFAULT_MODEL: str = ""
    BACKEND_NAME: BackendType

    def __init__(
        self,
        model: str = "",
        stream: bool = True,
        temperature: float | None | NotGiven = NOT_GIVEN,
        context_length_control: ContextLengthControlType = defs.CONTEXT_LENGTH_CONTROL,
        random_endpoint: bool = True,
        endpoint_id: str = "",
        http_client: httpx.Client | None = None,
        backend_name: str | None = None,
        settings: "Settings | SettingsDict | None" = None,  # Use default settings if not provided
    ):
        super().__init__(
            model,
            stream,
            temperature,
            context_length_control,
            random_endpoint,
            endpoint_id,
            http_client,
            backend_name,
            settings,
        )
        self.model_id = None
        self.endpoint = None

    @cached_property
    def raw_client(self) -> OpenAI | AzureOpenAI:
        self.endpoint, self.model_id = self._set_endpoint()

        if self.endpoint.proxy and self.http_client is None:
            self.http_client = httpx.Client(proxy=self.endpoint.proxy)

        if self.endpoint.is_azure or self.endpoint.endpoint_type == "openai_azure":
            if self.endpoint.api_base is None:
                raise ValueError("Azure endpoint is not set")
            return AzureOpenAI(
                azure_endpoint=self.endpoint.api_base,
                api_key=self.endpoint.api_key,
                api_version="2025-04-01-preview",
                http_client=self.http_client,
            )
        elif self.endpoint.endpoint_type == "openai_vertex":
            if self.endpoint.credentials is None:
                raise ValueError("OpenAI Vertex endpoint requires credentials")
            access_token, expires_at = get_token_with_cache(
                credentials=self.endpoint.credentials,
                proxy=self.endpoint.proxy,
                cached_token=self.endpoint.access_token,
                cached_expires_at=self.endpoint.access_token_expires_at,
            )
            self.endpoint.access_token = access_token
            self.endpoint.access_token_expires_at = expires_at
            return OpenAI(
                api_key=access_token,
                base_url=self.endpoint.api_base,
                http_client=self.http_client,
            )
        else:
            return OpenAI(
                api_key=self.endpoint.api_key,
                base_url=self.endpoint.api_base,
                http_client=self.http_client,
            )

    @overload
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage:
        pass

    @overload
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[True],
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> Generator[ChatCompletionDeltaMessage, Any, None]:
        pass

    @overload
    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: bool,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage | Generator[ChatCompletionDeltaMessage, Any, None]:
        pass

    def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] | Literal[True] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ):
        if model is not None:
            self.model = model
        if stream is not None:
            self.stream = stream
        if isinstance(temperature, AnthropicNotGiven):
            temperature = NOT_GIVEN
        if temperature is not None:
            self.temperature = temperature
        if isinstance(top_p, AnthropicNotGiven):
            top_p = NOT_GIVEN
        if isinstance(max_tokens, AnthropicNotGiven):
            max_tokens = NOT_GIVEN
        if isinstance(stream_options, AnthropicNotGiven):
            stream_options = NOT_GIVEN

        raw_client = self.raw_client  # 调用完 self.raw_client 后，self.model_id 会被赋值
        self.model_setting = self.backend_settings.models[self.model]
        if self.model_id is None:
            self.model_id = self.model_setting.id

        is_gemini3 = is_gemini_3_model(self.model, self.backend_name)

        # Get thinking tags for the current backend
        start_tag, end_tag = get_thinking_tags(self.backend_name)

        if not skip_cutoff and self.context_length_control == ContextLengthControlType.Latest:
            messages = cutoff_messages(
                messages,
                max_count=self.model_setting.context_length,
                backend=self.BACKEND_NAME,
                model=self.model,
            )

        if tools:
            if self.model_setting.function_call_available:
                _tools = tools
                if self.BACKEND_NAME.value == BackendType.MiniMax.value:  # MiniMax 就非要搞特殊
                    _tools = []
                    for tool in tools:
                        _tools.append(
                            {
                                "type": "function",
                                "function": {
                                    "name": tool["function"]["name"],
                                    "description": tool["function"].get("description", ""),
                                    "parameters": json.dumps(tool["function"].get("parameters", {})),
                                },
                            }
                        )
                tools_params = {"tools": _tools, "tool_choice": tool_choice}
            else:
                tools_str = json.dumps(tools, ensure_ascii=False, indent=None)
                additional_system_prompt = generate_tool_use_system_prompt(tools=tools_str)
                if messages and messages[0].get("role") == "system":
                    messages[0]["content"] += "\n\n" + additional_system_prompt
                else:
                    messages.insert(0, {"role": "system", "content": additional_system_prompt})
                tools_params = {}
        else:
            tools_params = {}

        # 只有 None 的时候才自动计算，否则就设置为 NOT_GIVEN
        if max_tokens is None and not max_completion_tokens:
            max_output_tokens = self.model_setting.max_output_tokens
            native_multimodal = self.model_setting.native_multimodal
            token_counts = get_message_token_counts(messages=messages, tools=tools, model=self.model, native_multimodal=native_multimodal)
            if max_output_tokens is not None:
                max_tokens = self.model_setting.context_length - token_counts - 64
                max_tokens = min(max(max_tokens, 1), max_output_tokens)
            else:
                max_tokens = self.model_setting.context_length - token_counts - 64

        if "o3-mini" in self.model_id or "o4-mini" in self.model_id or "gpt-5" in self.model_id:
            max_completion_tokens = max_tokens
            max_tokens = NOT_GIVEN

        if self.endpoint is None:
            raise ValueError("Endpoint is not set")
        endpoint = self.endpoint
        model_id = cast(str, self.model_id)
        raw_client = cast(Any, raw_client)
        self._acquire_rate_limit(endpoint, self.model, messages)

        # When endpoint is configured to use the new Responses API
        if endpoint.response_api:
            # Compute max_output_tokens for Responses API
            req_max_output_tokens = None
            if max_completion_tokens is not NOT_GIVEN and max_completion_tokens is not None:
                req_max_output_tokens = cast(int, max_completion_tokens)
            elif max_tokens is not NOT_GIVEN and max_tokens is not None:
                req_max_output_tokens = cast(int, max_tokens)

            if tools_params.get("tools", []):
                tools_params["tools"] = [tool_schema_for_response_api(tool) for tool in cast(list[Any], tools_params["tools"])]

            if self.stream:
                # Streaming via Responses API
                def generator():
                    stream_ctx = raw_client.responses.stream(
                        model=model_id,
                        input=messages,
                        temperature=None if self.temperature is NOT_GIVEN else self.temperature,
                        top_p=None if top_p is NOT_GIVEN else top_p,
                        max_output_tokens=req_max_output_tokens,
                        **tools_params,
                    )

                    usage = None
                    final_tool_calls: dict[int, dict] = {}
                    with stream_ctx as stream_response:
                        for event in stream_response:
                            messages_to_yield, usage_from_event = adapt_response_api_stream_event(
                                event=event,
                                final_tool_calls=final_tool_calls,
                                is_gemini3=is_gemini3,
                            )
                            if usage_from_event is not None:
                                usage = usage_from_event
                            yield from messages_to_yield

                        if usage:
                            yield ChatCompletionDeltaMessage(usage=usage)

                return generator()
            else:
                # Non-streaming via Responses API
                response = raw_client.responses.create(
                    model=model_id,
                    input=messages,
                    temperature=None if self.temperature is NOT_GIVEN else self.temperature,
                    top_p=None if top_p is NOT_GIVEN else top_p,
                    max_output_tokens=req_max_output_tokens,
                    **tools_params,
                )

                # Build ChatCompletion-like message
                content_text = response.output_text
                tool_calls = []
                for item in response.output or []:
                    if item.type == "function_call":
                        tool_calls.append(
                            {
                                "id": item.call_id or item.id,
                                "type": "function",
                                "function": {
                                    "name": item.name,
                                    "arguments": item.arguments,
                                },
                            }
                        )

                usage = None
                if response.usage:
                    u = response.usage
                    usage = Usage(
                        completion_tokens=u.output_tokens or 0,
                        prompt_tokens=u.input_tokens or 0,
                        total_tokens=(u.input_tokens or 0) + (u.output_tokens or 0),
                    )

                result = {
                    "content": content_text,
                    "tool_calls": tool_calls or None,
                    "usage": usage,
                }
                return ChatCompletionMessage(**cast(dict[str, Any], result))

        if self.stream:
            stream_response = raw_client.chat.completions.create(
                model=model_id,
                messages=messages,
                stream=True,
                temperature=self.temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                audio=audio,
                frequency_penalty=frequency_penalty,
                logit_bias=logit_bias,
                logprobs=logprobs,
                max_completion_tokens=max_completion_tokens,
                metadata=metadata,
                modalities=modalities,
                n=n,
                parallel_tool_calls=parallel_tool_calls,
                prediction=prediction,
                presence_penalty=presence_penalty,
                reasoning_effort=reasoning_effort,
                seed=seed,
                service_tier=service_tier,
                stop=stop,
                store=store,
                top_logprobs=top_logprobs,
                user=user,
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                stream_options=stream_options,
                response_format=response_format if response_format and self.model_setting.response_format_available else OPENAI_NOT_GIVEN,
                **tools_params,
            )

            def generator():
                full_content = ""
                result = {}
                usage = None
                buffer = ""
                in_reasoning = False
                accumulated_reasoning = []
                accumulated_content = []

                for chunk in stream_response:
                    if chunk.usage and chunk.usage.total_tokens:
                        if getattr(chunk.usage, "cached_tokens", None):  # noqa: B009
                            if getattr(chunk.usage, "prompt_tokens_details", None):  # noqa: B009
                                chunk.usage.prompt_tokens_details.cached_tokens = chunk.usage.cached_tokens
                            else:
                                chunk.usage.prompt_tokens_details = PromptTokensDetails(
                                    cached_tokens=chunk.usage.cached_tokens
                                )
                        usage = Usage(
                            completion_tokens=chunk.usage.completion_tokens or 0,
                            prompt_tokens=chunk.usage.prompt_tokens or 0,
                            total_tokens=chunk.usage.total_tokens or 0,
                            prompt_tokens_details=chunk.usage.prompt_tokens_details,
                            completion_tokens_details=chunk.usage.completion_tokens_details,
                        )

                    if not chunk.choices or len(chunk.choices) == 0 or not chunk.choices[0].delta:
                        if usage:
                            yield ChatCompletionDeltaMessage(usage=usage)
                        continue

                    if self.model_setting.function_call_available:
                        if chunk.choices[0].delta.tool_calls:
                            for index, tool_call in enumerate(chunk.choices[0].delta.tool_calls):
                                tool_call.index = index
                                tool_call.type = "function"  # 也是 MiniMax 的不规范导致的问题

                        # 即使支持 function call，也要处理思考标签
                        message = chunk.choices[0].delta.model_dump()
                        if "reasoning" in message and not message.get("reasoning_content"):
                            message["reasoning_content"] = message["reasoning"]
                        if is_gemini3 and chunk.choices[0].delta.tool_calls:
                            google_extras: list[dict] = []
                            for tool_call in chunk.choices[0].delta.tool_calls:
                                extra_content = getattr(tool_call, "extra_content", None)
                                if isinstance(extra_content, dict) and extra_content.get("google"):
                                    google_extras.append(extra_content["google"])
                            if google_extras:
                                message["raw_content"] = {"google": google_extras}

                        # Check for Gemini-3 thought content via extra_content
                        delta_extra_content = getattr(chunk.choices[0].delta, "extra_content", None)
                        is_thought_chunk = is_gemini3 and isinstance(delta_extra_content, dict) and delta_extra_content.get("google", {}).get("thought") is True

                        delta_content = message.get("content", "")
                        delta_reasoning = message.get("reasoning_content", "")
                        has_reasoning_output = False
                        if delta_reasoning:
                            accumulated_reasoning.append(delta_reasoning)
                            message["reasoning_content"] = delta_reasoning
                            if not delta_content:
                                message["content"] = ""
                            has_reasoning_output = True

                        if delta_content:
                            if is_thought_chunk:
                                # Gemini-3 thought content via extra_content, treat as reasoning directly
                                accumulated_reasoning.append(delta_content)
                                message["reasoning_content"] = delta_content
                                message["content"] = ""
                                has_reasoning_output = True
                            else:
                                buffer += delta_content

                                # 处理缓冲区中的内容，提取思考标签
                                buffer, current_output_content, current_reasoning_content, in_reasoning = process_thinking_content(buffer, in_reasoning, start_tag, end_tag)

                                # 累积内容
                                if current_output_content:
                                    accumulated_content.append(current_output_content)
                                if current_reasoning_content:
                                    accumulated_reasoning.append(current_reasoning_content)

                                # 只要有内容变化就产生 delta
                                if current_output_content or current_reasoning_content:
                                    if current_output_content:
                                        message["content"] = current_output_content
                                    elif current_reasoning_content:
                                        message["reasoning_content"] = current_reasoning_content
                                        message["content"] = ""  # 推理时不输出普通内容
                                        has_reasoning_output = True
                                else:
                                    # 处理后没有产生任何内容（如仅包含 </thought> 标签）
                                    # 清空 content 避免标签泄露到输出
                                    message["content"] = ""
                                    if not message.get("tool_calls") and not has_reasoning_output:
                                        # 如果没有任何内容且没有 tool_calls，则跳过这个消息
                                        continue

                        yield ChatCompletionDeltaMessage(**message, usage=usage)
                    else:
                        message = chunk.choices[0].delta.model_dump()
                        if "reasoning" in message and not message.get("reasoning_content"):
                            message["reasoning_content"] = message["reasoning"]

                        # Check for Gemini-3 thought content via extra_content
                        delta_extra_content = getattr(chunk.choices[0].delta, "extra_content", None)
                        is_thought_chunk = is_gemini3 and isinstance(delta_extra_content, dict) and delta_extra_content.get("google", {}).get("thought") is True

                        delta_content = message.get("content", "")
                        delta_reasoning = message.get("reasoning_content", "")
                        has_reasoning_output = False
                        yielded = False
                        if delta_reasoning:
                            accumulated_reasoning.append(delta_reasoning)
                            message["reasoning_content"] = delta_reasoning
                            if not delta_content:
                                message["content"] = ""
                                yield ChatCompletionDeltaMessage(**message, usage=usage)
                                continue
                            has_reasoning_output = True

                        if delta_content:
                            if is_thought_chunk:
                                # Gemini-3 thought content via extra_content, treat as reasoning directly
                                accumulated_reasoning.append(delta_content)
                                message["reasoning_content"] = delta_content
                                message["content"] = ""
                                yield ChatCompletionDeltaMessage(**message, usage=usage)
                            else:
                                buffer += delta_content

                                # 处理缓冲区中的内容，提取思考标签
                                buffer, current_output_content, current_reasoning_content, in_reasoning = process_thinking_content(buffer, in_reasoning, start_tag, end_tag)

                                # 累积内容
                                if current_output_content:
                                    accumulated_content.append(current_output_content)
                                if current_reasoning_content:
                                    accumulated_reasoning.append(current_reasoning_content)
                                    has_reasoning_output = True

                                # 只要有内容变化就产生 delta
                                if current_output_content or current_reasoning_content:
                                    if current_output_content:
                                        message["content"] = current_output_content
                                    elif current_reasoning_content:
                                        message["reasoning_content"] = current_reasoning_content
                                        message["content"] = ""  # 推理时不输出普通内容

                                    if tools:
                                        full_content += current_output_content
                                        tool_call_data = ToolCallContentProcessor(full_content).tool_calls
                                        if tool_call_data:
                                            message["tool_calls"] = tool_call_data["tool_calls"]

                                    if full_content in ("<", "<|", "<|▶", "<|▶|") or full_content.startswith("<|▶|>"):
                                        message["content"] = ""
                                        result = message
                                        continue

                                    yield ChatCompletionDeltaMessage(**message, usage=usage)
                                    yielded = True
                                elif delta_content and not has_reasoning_output:
                                    # 有原始内容但处理后没有产生任何内容（如仅包含 </thought> 标签）
                                    # 清空并跳过，避免标签泄露到输出
                                    pass
                                elif delta_reasoning and not yielded:
                                    # 有 reasoning_content，但内容处理后没有产生输出（如仅包含结束标签）
                                    yield ChatCompletionDeltaMessage(**message, usage=usage)

                # 处理最后剩余的缓冲区内容
                if buffer:
                    if in_reasoning:
                        accumulated_reasoning.append(buffer)
                    else:
                        accumulated_content.append(buffer)

                    final_message = {}
                    if accumulated_content:
                        final_content = "".join(accumulated_content)
                        if final_content.strip():  # 只有当内容非空时才输出
                            final_message["content"] = final_content

                    if accumulated_reasoning:
                        final_reasoning = "".join(accumulated_reasoning)
                        if final_reasoning.strip():  # 只有当推理内容非空时才输出
                            final_message["reasoning_content"] = final_reasoning

                    if final_message:
                        yield ChatCompletionDeltaMessage(**final_message, usage=usage)

                if result:
                    yield ChatCompletionDeltaMessage(**result, usage=usage)

            return generator()
        else:
            response = raw_client.chat.completions.create(
                model=model_id,
                messages=messages,
                stream=False,
                temperature=self.temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                audio=audio,
                frequency_penalty=frequency_penalty,
                logit_bias=logit_bias,
                logprobs=logprobs,
                max_completion_tokens=max_completion_tokens,
                metadata=metadata,
                modalities=modalities,
                n=n,
                parallel_tool_calls=parallel_tool_calls,
                prediction=prediction,
                presence_penalty=presence_penalty,
                reasoning_effort=reasoning_effort,
                seed=seed,
                service_tier=service_tier,
                stop=stop,
                store=store,
                top_logprobs=top_logprobs,
                user=user,
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                stream_options=stream_options,
                response_format=response_format if response_format and self.model_setting.response_format_available else OPENAI_NOT_GIVEN,
                **tools_params,
            )

            if not response.choices:
                raise ValueError(f"No response choices: {response}")

            if response.usage and getattr(response.usage, "cached_tokens", None):  # noqa: B009
                if getattr(response.usage, "prompt_tokens_details", None):  # noqa: B009
                    response.usage.prompt_tokens_details.cached_tokens = response.usage.cached_tokens
                else:
                    response.usage.prompt_tokens_details = PromptTokensDetails(
                        cached_tokens=response.usage.cached_tokens
                    )

            message_obj = response.choices[0].message
            reasoning_content = getattr(message_obj, "reasoning_content", None)
            if reasoning_content is None:
                reasoning_content = getattr(message_obj, "reasoning", None)
            result = {
                "content": message_obj.content,
                "reasoning_content": reasoning_content,
                "usage": response.usage.model_dump() if response.usage else None,
            }

            if not result["reasoning_content"] and result["content"]:
                result["content"], extracted_reasoning = extract_reasoning_tagged_content(result["content"])
                if extracted_reasoning:
                    result["reasoning_content"] = extracted_reasoning

            if tools:
                if self.model_setting.function_call_available:
                    if response.choices[0].message.tool_calls:
                        result["tool_calls"] = [{**tool_call.model_dump(), "type": "function"} for tool_call in response.choices[0].message.tool_calls]
                else:
                    tool_call_content_processor = ToolCallContentProcessor(result["content"])
                    tool_call_data = tool_call_content_processor.tool_calls
                    if tool_call_data:
                        result["tool_calls"] = tool_call_data["tool_calls"]
                        result["content"] = tool_call_content_processor.non_tool_content

            # 对 Gemini 3 模型保留 Google 扩展信息（如 thought_signature），以便上层可以按文档转发
            if is_gemini3 and getattr(response.choices[0].message, "tool_calls", None):
                google_extras: list[dict] = []
                for tool_call in response.choices[0].message.tool_calls:
                    extra_content = getattr(tool_call, "extra_content", None)
                    if isinstance(extra_content, dict) and extra_content.get("google"):
                        google_extras.append(extra_content["google"])
                if google_extras:
                    result["raw_content"] = [{"google": google_extras}]

            return ChatCompletionMessage(**cast(dict[str, Any], result))


class AsyncOpenAICompatibleChatClient(BaseAsyncChatClient):
    DEFAULT_MODEL: str = ""
    BACKEND_NAME: BackendType

    def __init__(
        self,
        model: str = "",
        stream: bool = True,
        temperature: float | None | NotGiven = NOT_GIVEN,
        context_length_control: ContextLengthControlType = defs.CONTEXT_LENGTH_CONTROL,
        random_endpoint: bool = True,
        endpoint_id: str = "",
        http_client: httpx.AsyncClient | None = None,
        backend_name: str | None = None,
        settings: "Settings | SettingsDict | None" = None,  # Use default settings if not provided
    ):
        super().__init__(
            model,
            stream,
            temperature,
            context_length_control,
            random_endpoint,
            endpoint_id,
            http_client,
            backend_name,
            settings,
        )
        self.model_id = None
        self.endpoint = None

    @cached_property
    def raw_client(self):
        self.endpoint, self.model_id = self._set_endpoint()

        if self.endpoint.proxy and self.http_client is None:
            self.http_client = httpx.AsyncClient(proxy=self.endpoint.proxy)

        if self.endpoint.is_azure or self.endpoint.endpoint_type == "openai_azure":
            if self.endpoint.api_base is None:
                raise ValueError("Azure endpoint is not set")
            return AsyncAzureOpenAI(
                azure_endpoint=self.endpoint.api_base,
                api_key=self.endpoint.api_key,
                api_version="2025-04-01-preview",
                http_client=self.http_client,
            )
        elif self.endpoint.endpoint_type == "openai_vertex":
            if self.endpoint.credentials is None:
                raise ValueError("OpenAI Vertex endpoint requires credentials")
            access_token, expires_at = get_token_with_cache(
                credentials=self.endpoint.credentials,
                proxy=self.endpoint.proxy,
                cached_token=self.endpoint.access_token,
                cached_expires_at=self.endpoint.access_token_expires_at,
            )
            self.endpoint.access_token = access_token
            self.endpoint.access_token_expires_at = expires_at
            return AsyncOpenAI(
                api_key=access_token,
                base_url=self.endpoint.api_base,
                http_client=self.http_client,
            )
        else:
            return AsyncOpenAI(
                api_key=self.endpoint.api_key,
                base_url=self.endpoint.api_base,
                http_client=self.http_client,
            )

    @overload
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage:
        pass

    @overload
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[True],
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        pass

    @overload
    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: bool,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage | AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        pass

    async def create_completion(
        self,
        *,
        messages: list,
        model: str | None = None,
        stream: Literal[False] | Literal[True] = False,
        temperature: float | None | NotGiven = NOT_GIVEN,
        max_tokens: int | None | NotGiven = None,
        tools: Iterable[ToolParam] | NotGiven = NOT_GIVEN,
        tool_choice: ToolChoice | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        stream_options: ChatCompletionStreamOptionsParam | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven | None = NOT_GIVEN,
        skip_cutoff: bool = False,
        audio: ChatCompletionAudioParam | OpenAINotGiven | None = NOT_GIVEN,
        frequency_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        logit_bias: dict[str, int] | OpenAINotGiven | None = NOT_GIVEN,
        logprobs: bool | OpenAINotGiven | None = NOT_GIVEN,
        max_completion_tokens: int | OpenAINotGiven | None = NOT_GIVEN,
        metadata: Metadata | OpenAINotGiven | None = NOT_GIVEN,
        modalities: list[ChatCompletionModality] | OpenAINotGiven | None = NOT_GIVEN,
        n: int | OpenAINotGiven | None = NOT_GIVEN,
        parallel_tool_calls: bool | OpenAINotGiven = NOT_GIVEN,
        prediction: ChatCompletionPredictionContentParam | OpenAINotGiven | None = NOT_GIVEN,
        presence_penalty: float | OpenAINotGiven | None = NOT_GIVEN,
        reasoning_effort: ChatCompletionReasoningEffort | OpenAINotGiven | None = NOT_GIVEN,
        thinking: ThinkingConfigParam | None | NotGiven = NOT_GIVEN,
        seed: int | OpenAINotGiven | None = NOT_GIVEN,
        service_tier: Literal["auto", "default"] | OpenAINotGiven | None = NOT_GIVEN,
        stop: str | list[str] | OpenAINotGiven | None = NOT_GIVEN,
        store: bool | OpenAINotGiven | None = NOT_GIVEN,
        top_logprobs: int | OpenAINotGiven | None = NOT_GIVEN,
        user: str | OpenAINotGiven = NOT_GIVEN,
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | OpenAINotGiven = NOT_GIVEN,
    ) -> ChatCompletionMessage | AsyncGenerator[ChatCompletionDeltaMessage, Any]:
        if model is not None:
            self.model = model
        if stream is not None:
            self.stream = stream
        if isinstance(temperature, AnthropicNotGiven):
            temperature = NOT_GIVEN
        if temperature is not None:
            self.temperature = temperature
        if isinstance(top_p, AnthropicNotGiven):
            top_p = NOT_GIVEN
        if isinstance(max_tokens, AnthropicNotGiven):
            max_tokens = NOT_GIVEN
        if isinstance(stream_options, AnthropicNotGiven):
            stream_options = NOT_GIVEN

        raw_client = self.raw_client  # 调用完 self.raw_client 后，self.model_id 会被赋值
        self.model_setting = self.backend_settings.models[self.model]
        if self.model_id is None:
            self.model_id = self.model_setting.id

        is_gemini3 = is_gemini_3_model(self.model, self.backend_name)

        # Get thinking tags for the current backend
        start_tag, end_tag = get_thinking_tags(self.backend_name)

        if not skip_cutoff and self.context_length_control == ContextLengthControlType.Latest:
            messages = cutoff_messages(
                messages,
                max_count=self.model_setting.context_length,
                backend=self.BACKEND_NAME,
                model=self.model,
            )

        if tools:
            if self.model_setting.function_call_available:
                _tools = tools
                if self.BACKEND_NAME.value == BackendType.MiniMax.value:
                    _tools = []
                    for tool in tools:
                        _tools.append(
                            {
                                "type": "function",
                                "function": {
                                    "name": tool["function"]["name"],
                                    "description": tool["function"].get("description", ""),
                                    "parameters": json.dumps(tool["function"].get("parameters", {})),
                                },
                            }
                        )
                tools_params = {"tools": _tools, "tool_choice": tool_choice}
            else:
                tools_str = json.dumps(tools, ensure_ascii=False, indent=None)
                additional_system_prompt = generate_tool_use_system_prompt(tools=tools_str)
                if messages and messages[0].get("role") == "system":
                    messages[0]["content"] += "\n\n" + additional_system_prompt
                else:
                    messages.insert(0, {"role": "system", "content": additional_system_prompt})
                tools_params = {}
        else:
            tools_params = {}

        # 只有 None 的时候才自动计算，否则就设置为 NOT_GIVEN
        if max_tokens is None and not max_completion_tokens:
            max_output_tokens = self.model_setting.max_output_tokens
            native_multimodal = self.model_setting.native_multimodal
            token_counts = get_message_token_counts(messages=messages, tools=tools, model=self.model, native_multimodal=native_multimodal)
            if max_output_tokens is not None:
                max_tokens = self.model_setting.context_length - token_counts - 64
                max_tokens = min(max(max_tokens, 1), max_output_tokens)
            else:
                max_tokens = self.model_setting.context_length - token_counts - 64

        if "o3-mini" in self.model_id or "o4-mini" in self.model_id or "gpt-5" in self.model_id:
            max_completion_tokens = max_tokens
            max_tokens = NOT_GIVEN

        if self.endpoint is None:
            raise ValueError("Endpoint is not set")
        endpoint = self.endpoint
        model_id = cast(str, self.model_id)
        raw_client = cast(Any, raw_client)
        await self._acquire_rate_limit(endpoint, self.model, messages)

        # When endpoint is configured to use the new Responses API
        if endpoint.response_api:
            req_max_output_tokens = None
            if max_completion_tokens is not NOT_GIVEN and max_completion_tokens is not None:
                req_max_output_tokens = cast(int, max_completion_tokens)
            elif max_tokens is not NOT_GIVEN and max_tokens is not None:
                req_max_output_tokens = cast(int, max_tokens)

            if tools_params.get("tools", []):
                try:
                    tools_params["tools"] = [tool_schema_for_response_api(tool) for tool in cast(list[Any], tools_params["tools"])]
                except Exception:
                    pass

            if self.stream:

                async def _generator():
                    usage = None
                    final_tool_calls: dict[int, dict] = {}
                    try:
                        # Async streaming via Responses API
                        async with raw_client.responses.stream(
                            model=model_id,
                            input=messages,
                            temperature=None if self.temperature is NOT_GIVEN else self.temperature,
                            top_p=None if top_p is NOT_GIVEN else top_p,
                            max_output_tokens=req_max_output_tokens,
                            **tools_params,
                        ) as stream_response:
                            async for event in stream_response:
                                messages_to_yield, usage_from_event = adapt_response_api_stream_event(
                                    event=event,
                                    final_tool_calls=final_tool_calls,
                                    is_gemini3=is_gemini3,
                                )
                                if usage_from_event is not None:
                                    usage = usage_from_event
                                for message in messages_to_yield:
                                    yield message

                    finally:
                        if usage:
                            yield ChatCompletionDeltaMessage(usage=usage)

                return _generator()
            else:
                response = await raw_client.responses.create(
                    model=model_id,
                    input=messages,
                    temperature=None if self.temperature is NOT_GIVEN else self.temperature,
                    top_p=None if top_p is NOT_GIVEN else top_p,
                    max_output_tokens=req_max_output_tokens,
                    **tools_params,
                )

                content_text = response.output_text
                tool_calls = []
                for item in response.output or []:
                    if item.type == "function_call":
                        tool_calls.append(
                            {
                                "id": item.call_id or item.id,
                                "type": "function",
                                "function": {
                                    "name": item.name,
                                    "arguments": item.arguments,
                                },
                            }
                        )

                usage = None
                if response.usage:
                    u = response.usage
                    usage = Usage(
                        completion_tokens=u.output_tokens or 0,
                        prompt_tokens=u.input_tokens or 0,
                        total_tokens=(u.input_tokens or 0) + (u.output_tokens or 0),
                    )

                result = {
                    "content": content_text,
                    "tool_calls": tool_calls or None,
                    "usage": usage,
                }
                return ChatCompletionMessage(**cast(dict[str, Any], result))

        if self.stream:
            stream_response = await raw_client.chat.completions.create(
                model=model_id,
                messages=messages,
                stream=self.stream,
                temperature=self.temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                audio=audio,
                frequency_penalty=frequency_penalty,
                logit_bias=logit_bias,
                logprobs=logprobs,
                max_completion_tokens=max_completion_tokens,
                metadata=metadata,
                modalities=modalities,
                n=n,
                parallel_tool_calls=parallel_tool_calls,
                prediction=prediction,
                presence_penalty=presence_penalty,
                reasoning_effort=reasoning_effort,
                seed=seed,
                service_tier=service_tier,
                stop=stop,
                store=store,
                top_logprobs=top_logprobs,
                user=user,
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                stream_options=stream_options,
                response_format=response_format if response_format and self.model_setting.response_format_available else OPENAI_NOT_GIVEN,
                **tools_params,
            )

            async def _generator():
                full_content = ""
                result = {}
                usage = None
                buffer = ""
                in_reasoning = False
                accumulated_reasoning = []
                accumulated_content = []

                async for chunk in stream_response:
                    if chunk.usage and chunk.usage.total_tokens:
                        if getattr(chunk.usage, "cached_tokens", None):  # noqa: B009
                            if getattr(chunk.usage, "prompt_tokens_details", None):  # noqa: B009
                                chunk.usage.prompt_tokens_details.cached_tokens = chunk.usage.cached_tokens
                            else:
                                chunk.usage.prompt_tokens_details = PromptTokensDetails(
                                    cached_tokens=chunk.usage.cached_tokens
                                )
                        usage = Usage(
                            completion_tokens=chunk.usage.completion_tokens or 0,
                            prompt_tokens=chunk.usage.prompt_tokens or 0,
                            total_tokens=chunk.usage.total_tokens or 0,
                            completion_tokens_details=chunk.usage.completion_tokens_details,
                            prompt_tokens_details=chunk.usage.prompt_tokens_details,
                        )

                    if not chunk.choices or len(chunk.choices) == 0 or not chunk.choices[0].delta:
                        if usage:
                            yield ChatCompletionDeltaMessage(usage=usage)
                        continue

                    if self.model_setting.function_call_available:
                        if chunk.choices[0].delta.tool_calls:
                            for index, tool_call in enumerate(chunk.choices[0].delta.tool_calls):
                                tool_call.index = index
                                tool_call.type = "function"

                        # 即使支持 function call，也要处理思考标签
                        message = chunk.choices[0].delta.model_dump()
                        if "reasoning" in message and not message.get("reasoning_content"):
                            message["reasoning_content"] = message["reasoning"]
                        if is_gemini3 and chunk.choices[0].delta.tool_calls:
                            google_extras: list[dict] = []
                            for tool_call in chunk.choices[0].delta.tool_calls:
                                extra_content = getattr(tool_call, "extra_content", None)
                                if isinstance(extra_content, dict) and extra_content.get("google"):
                                    google_extras.append(extra_content["google"])
                            if google_extras:
                                message["raw_content"] = {"google": google_extras}

                        # Check for Gemini-3 thought content via extra_content
                        delta_extra_content = getattr(chunk.choices[0].delta, "extra_content", None)
                        is_thought_chunk = is_gemini3 and isinstance(delta_extra_content, dict) and delta_extra_content.get("google", {}).get("thought") is True

                        delta_content = message.get("content", "")
                        delta_reasoning = message.get("reasoning_content", "")
                        has_reasoning_output = False
                        if delta_reasoning:
                            accumulated_reasoning.append(delta_reasoning)
                            message["reasoning_content"] = delta_reasoning
                            if not delta_content:
                                message["content"] = ""
                            has_reasoning_output = True

                        if delta_content:
                            if is_thought_chunk:
                                # Gemini-3 thought content via extra_content, treat as reasoning directly
                                accumulated_reasoning.append(delta_content)
                                message["reasoning_content"] = delta_content
                                message["content"] = ""
                                has_reasoning_output = True
                            else:
                                buffer += delta_content

                                # 处理缓冲区中的内容，提取思考标签
                                buffer, current_output_content, current_reasoning_content, in_reasoning = process_thinking_content(buffer, in_reasoning, start_tag, end_tag)

                                # 累积内容
                                if current_output_content:
                                    accumulated_content.append(current_output_content)
                                if current_reasoning_content:
                                    accumulated_reasoning.append(current_reasoning_content)

                                # 只要有内容变化就产生 delta
                                if current_output_content or current_reasoning_content:
                                    if current_output_content:
                                        message["content"] = current_output_content
                                    elif current_reasoning_content:
                                        message["reasoning_content"] = current_reasoning_content
                                        message["content"] = ""  # 推理时不输出普通内容
                                        has_reasoning_output = True
                                else:
                                    # 处理后没有产生任何内容（如仅包含 </thought> 标签）
                                    # 清空 content 避免标签泄露到输出
                                    message["content"] = ""
                                    if not message.get("tool_calls") and not has_reasoning_output:
                                        # 如果没有任何内容且没有 tool_calls，则跳过这个消息
                                        continue

                        yield ChatCompletionDeltaMessage(**message, usage=usage)
                    else:
                        message = chunk.choices[0].delta.model_dump()
                        if "reasoning" in message and not message.get("reasoning_content"):
                            message["reasoning_content"] = message["reasoning"]

                        # Check for Gemini-3 thought content via extra_content
                        delta_extra_content = getattr(chunk.choices[0].delta, "extra_content", None)
                        is_thought_chunk = is_gemini3 and isinstance(delta_extra_content, dict) and delta_extra_content.get("google", {}).get("thought") is True

                        delta_content = message.get("content", "")
                        delta_reasoning = message.get("reasoning_content", "")
                        has_reasoning_output = False
                        yielded = False
                        if delta_reasoning:
                            accumulated_reasoning.append(delta_reasoning)
                            message["reasoning_content"] = delta_reasoning
                            if not delta_content:
                                message["content"] = ""
                                yield ChatCompletionDeltaMessage(**message, usage=usage)
                                continue
                            has_reasoning_output = True

                        if delta_content:
                            if is_thought_chunk:
                                # Gemini-3 thought content via extra_content, treat as reasoning directly
                                accumulated_reasoning.append(delta_content)
                                message["reasoning_content"] = delta_content
                                message["content"] = ""
                                yield ChatCompletionDeltaMessage(**message, usage=usage)
                            else:
                                buffer += delta_content

                                # 处理缓冲区中的内容，提取思考标签
                                buffer, current_output_content, current_reasoning_content, in_reasoning = process_thinking_content(buffer, in_reasoning, start_tag, end_tag)

                                # 累积内容
                                if current_output_content:
                                    accumulated_content.append(current_output_content)
                                if current_reasoning_content:
                                    accumulated_reasoning.append(current_reasoning_content)
                                    has_reasoning_output = True

                                # 只要有内容变化就产生 delta
                                if current_output_content or current_reasoning_content:
                                    if current_output_content:
                                        message["content"] = current_output_content
                                    elif current_reasoning_content:
                                        message["reasoning_content"] = current_reasoning_content
                                        message["content"] = ""  # 推理时不输出普通内容

                                    if tools:
                                        full_content += current_output_content
                                        tool_call_data = ToolCallContentProcessor(full_content).tool_calls
                                        if tool_call_data:
                                            message["tool_calls"] = tool_call_data["tool_calls"]

                                    if full_content in ("<", "<|", "<|▶", "<|▶|") or full_content.startswith("<|▶|>"):
                                        message["content"] = ""
                                        result = message
                                        continue

                                    yield ChatCompletionDeltaMessage(**message, usage=usage)
                                    yielded = True
                                elif delta_content and not has_reasoning_output:
                                    # 有原始内容但处理后没有产生任何内容（如仅包含 </thought> 标签）
                                    # 清空并跳过，避免标签泄露到输出
                                    pass
                                elif delta_reasoning and not yielded:
                                    # 有 reasoning_content，但内容处理后没有产生输出（如仅包含结束标签）
                                    yield ChatCompletionDeltaMessage(**message, usage=usage)

                # 处理最后剩余的缓冲区内容
                if buffer:
                    if in_reasoning:
                        accumulated_reasoning.append(buffer)
                    else:
                        accumulated_content.append(buffer)

                    final_message = {}
                    if accumulated_content:
                        final_content = "".join(accumulated_content)
                        if final_content.strip():  # 只有当内容非空时才输出
                            final_message["content"] = final_content

                    if accumulated_reasoning:
                        final_reasoning = "".join(accumulated_reasoning)
                        if final_reasoning.strip():  # 只有当推理内容非空时才输出
                            final_message["reasoning_content"] = final_reasoning

                    if final_message:
                        yield ChatCompletionDeltaMessage(**final_message, usage=usage)

                if result:
                    yield ChatCompletionDeltaMessage(**result, usage=usage)

            return _generator()
        else:
            response = await raw_client.chat.completions.create(
                model=model_id,
                messages=messages,
                stream=self.stream,
                temperature=self.temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                audio=audio,
                frequency_penalty=frequency_penalty,
                logit_bias=logit_bias,
                logprobs=logprobs,
                max_completion_tokens=max_completion_tokens,
                metadata=metadata,
                modalities=modalities,
                n=n,
                parallel_tool_calls=parallel_tool_calls,
                prediction=prediction,
                presence_penalty=presence_penalty,
                reasoning_effort=reasoning_effort,
                seed=seed,
                service_tier=service_tier,
                stop=stop,
                store=store,
                top_logprobs=top_logprobs,
                user=user,
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                stream_options=stream_options,
                response_format=response_format if response_format and self.model_setting.response_format_available else OPENAI_NOT_GIVEN,
                **tools_params,
            )

            if not response.choices:
                raise ValueError(f"No response choices: {response}")

            if response.usage and getattr(response.usage, "cached_tokens", None):  # noqa: B009
                if getattr(response.usage, "prompt_tokens_details", None):  # noqa: B009
                    response.usage.prompt_tokens_details.cached_tokens = response.usage.cached_tokens
                else:
                    response.usage.prompt_tokens_details = PromptTokensDetails(
                        cached_tokens=response.usage.cached_tokens
                    )

            message_obj = response.choices[0].message
            reasoning_content = getattr(message_obj, "reasoning_content", None)
            if reasoning_content is None:
                reasoning_content = getattr(message_obj, "reasoning", None)
            result = {
                "content": message_obj.content,
                "reasoning_content": reasoning_content,
                "usage": response.usage.model_dump() if response.usage else None,
            }

            if not result["reasoning_content"] and result["content"]:
                result["content"], extracted_reasoning = extract_reasoning_tagged_content(result["content"])
                if extracted_reasoning:
                    result["reasoning_content"] = extracted_reasoning

            if tools:
                if self.model_setting.function_call_available:
                    if response.choices[0].message.tool_calls:
                        result["tool_calls"] = [{**tool_call.model_dump(), "type": "function"} for tool_call in response.choices[0].message.tool_calls]
                else:
                    tool_call_content_processor = ToolCallContentProcessor(result["content"])
                    tool_call_data = tool_call_content_processor.tool_calls
                    if tool_call_data:
                        result["tool_calls"] = tool_call_data["tool_calls"]
                        result["content"] = tool_call_content_processor.non_tool_content

            # 对 Gemini 3 模型保留 Google 扩展信息（如 thought_signature），以便上层可以按文档转发
            if is_gemini3 and getattr(response.choices[0].message, "tool_calls", None):
                google_extras: list[dict] = []
                for tool_call in response.choices[0].message.tool_calls:
                    extra_content = getattr(tool_call, "extra_content", None)
                    if isinstance(extra_content, dict) and extra_content.get("google"):
                        google_extras.append(extra_content["google"])
                if google_extras:
                    result["raw_content"] = [{"google": google_extras}]

            return ChatCompletionMessage(**cast(dict[str, Any], result))
